/*
 * Copyright 2019 European Commission
 *
 * Licensed under the EUPL, Version 1.2 or – as soon they will be approved by the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 *     https://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and limitations under the Licence.
 */
'use strict';

var groupsByOrganization = require('../group-organizations');
var orgFixtures = require('../../test/group-fixtures');

describe('group-organizations', function () {
  context ('when sorting organizations and their contained groups', function () {

    it ('should put the default organization groups last', function () {
      const defaultOrg = orgFixtures.defaultOrganization();
      const groups = [orgFixtures.expandedGroup({ organization: defaultOrg }),
                      orgFixtures.expandedGroup(),
                      orgFixtures.expandedGroup()];

      const sortedGroups = groupsByOrganization(groups);

      assert.equal(sortedGroups[2].organization.id, defaultOrg.id);
    });

    it ('should sort organizations by name', function () {
      const org1 = orgFixtures.organization({ name: 'zzzzz' });
      const org2 = orgFixtures.organization({ name: 'aaaaa' });
      const org3 = orgFixtures.organization({ name: 'yyyyy' });
      const groups = [orgFixtures.expandedGroup({ organization: org1 }),
                      orgFixtures.expandedGroup({ organization: org2 }),
                      orgFixtures.expandedGroup({ organization: org3 })];

      const sortedGroups = groupsByOrganization(groups);

      assert.equal(sortedGroups[0].organization.name, 'aaaaa');
      assert.equal(sortedGroups[1].organization.name, 'yyyyy');
      assert.equal(sortedGroups[2].organization.name, 'zzzzz');
    });

    it ('should sort organizations secondarily by id', function () {
      const org1 = orgFixtures.organization({ name: 'zzzzz', id: 'zzzzz' });
      const org2 = orgFixtures.organization({ name: 'zzzzz', id: 'aaaaa' });
      const org3 = orgFixtures.organization({ name: 'zzzzz', id: 'yyyyy' });
      const groups = [orgFixtures.expandedGroup({ organization: org1 }),
                      orgFixtures.expandedGroup({ organization: org2 }),
                      orgFixtures.expandedGroup({ organization: org3 })];

      const sortedGroups = groupsByOrganization(groups);

      assert.equal(sortedGroups[0].organization.id, 'aaaaa');
      assert.equal(sortedGroups[1].organization.id, 'yyyyy');
      assert.equal(sortedGroups[2].organization.id, 'zzzzz');
    });

    it ('should only include logo for first group in each organization', function () {
      const org = orgFixtures.organization({ name: 'Aluminum' });
      const org2 = orgFixtures.organization({ name: 'Zirconium' });
      const groups = [
        { name: 'Aluminum', organization: org },
        { name: 'Beryllium', organization: org2},
        { name: 'Butane', organization: org2 },
        { name: 'Cadmium', organization: org},
      ];

      const sortedGroups = groupsByOrganization(groups);

      assert.equal(sortedGroups[0].logo, org.logo);
      assert.equal(typeof sortedGroups[1].logo, 'undefined');
      assert.equal(sortedGroups[2].logo, org2.logo);
      assert.equal(typeof sortedGroups[3].logo, 'undefined');
    });
  });

  context('when encountering missing data', function () {

    it('should be able to sort without any groups in the default org', function () {
      const org = orgFixtures.organization({ name:'Aluminum' });
      const groups = [
        { name: 'Aluminum', organization: org },
        { name: 'Beryllium', organization: org},
        { name: 'Cadmium', organization: org},
      ];

      const sortedGroups = groupsByOrganization(groups);

      sortedGroups.forEach((group, idx) => {
        assert.equal(group.name, groups[idx].name);
      });

    });

    it('should omit any groups without an organization', function () {
      const org = orgFixtures.organization({ name:'Europium' });
      const groups = [
        { name: 'Aluminum', organization: org },
        { name: 'Beryllium', organization: org},
        { name: 'Butane' },
        { name: 'Cadmium', organization: org},
      ];

      const sortedGroups = groupsByOrganization(groups);

      assert.equal(sortedGroups.length, groups.length - 1);
      sortedGroups.forEach(group => {
        assert.notEqual(group.name, 'Butane');
      });
    });

    it('should omit any groups with unexpanded organizations', function () {
      const org = orgFixtures.organization({ name:'Europium' });
      const groups = [
        { name: 'Aluminum', organization: org },
        { name: 'Beryllium', organization: org},
        { name: 'Butane', organization: 'foobar' },
        { name: 'Cadmium', organization: org},
      ];

      const sortedGroups = groupsByOrganization(groups);

      assert.equal(sortedGroups.length, groups.length - 1);
      sortedGroups.forEach(group => {
        assert.notEqual(group.name, 'Butane');
      });

    });

    it ('should omit logo property if not present on organization', function () {
      const org = orgFixtures.organization({ logo: undefined });
      const org2 = orgFixtures.organization({ logo: null });
      const groups = [
        { name: 'Aluminum', organization: org },
        { name: 'Beryllium', organization: org2},
        { name: 'Butane', organization: org2 },
        { name: 'Cadmium', organization: org},
      ];

      const sortedGroups = groupsByOrganization(groups);

      sortedGroups.forEach(group => {
        assert.equal(typeof group.logo, 'undefined');
      });
    });
  });

  context('when building data structures', function () {

    it ('returned group objects should be immutable', function () {
      const group = orgFixtures.expandedGroup({name: 'Halfnium'});
      const groups = [group];

      const sortedGroups = groupsByOrganization(groups);
      assert.throws(() => {
        sortedGroups[0].name = 'Something Else';
      });
      assert.throws(() => {
        sortedGroups[0].links.html = 'dingdong';
      });

      assert.notEqual(group.name, 'Something Else');
      assert.notEqual(group.links.html, 'dingdong');
    });
  });
});
